#include <math.h>
#include <string.h>
#include "Python.h"
#define _ATOM3D_MODULE
#include "atom3d.h"
#define RADIANS_TO_DEGREES 57.29577951308232e0
#define DEGREES_TO_RADIANS 0.017453292519943295e0

static PyObject *ErrorObject;

#define ErrorSet(msg) PyErr_SetString(ErrorObject, msg)

Atom3d * Atom3d_Create(void)
{
	Atom3d *self = PyObject_NEW(Atom3d, &Atom3dType);
	if (self == NULL)
		return NULL;
	self->x = self->y = self->z = self->radius = 0.0e0;
	self->fp_distance = self->sp_angle = self->tp_torsion = 0.0e0;
	self->first_parent = self->second_parent = self->third_parent = NULL;
	strncpy(self->name,"    \0", 5);
	self->resnum = self->bsep0 = self->bsep1 = 0;
	self->sasa = self->sasarad = 0.0e0;
	return self;
}

double Atom3d_SqDistance(Atom3d *a1, Atom3d *a2)
{
	double dx = a1->x - a2->x;
	double dy = a1->y - a2->y;
	double dz = a1->z - a2->z;
	return (dx*dx + dy*dy + dz*dz);
}

double Atom3d_Distance(Atom3d *a1, Atom3d *a2)
{
	double dx = a1->x - a2->x;
	double dy = a1->y - a2->y;
	double dz = a1->z - a2->z;
	return sqrt(dx*dx + dy*dy + dz*dz);
}

double Atom3d_Angle(Atom3d *a1, Atom3d *a2, Atom3d *a3)
{
	double x1 = a1->x - a2->x;
	double y1 = a1->y - a2->y;
	double z1 = a1->z - a2->z;

	double x2 = a3->x - a2->x;
	double y2 = a3->y - a2->y;
	double z2 = a3->z - a2->z;

	double xx = y1*z2 - y2*z1;
	double yy = z1*x2 - z2*x1;
	double zz = x1*y2 - x2*y1;

	double cross = sqrt(xx*xx + yy*yy + zz*zz);
	double dot = x1*x2 + y1*y2 + z1*z2;

	double ang = fabs(atan2(cross, dot)) * RADIANS_TO_DEGREES;

	return ang;
}

double Atom3d_Torsion(Atom3d *a1, Atom3d *a2, Atom3d *a3, Atom3d *a4)
{
	double a = a1->x - a2->x; 
	double b = a1->y - a2->y; 
	double c = a1->z - a2->z;
	
	double d = a3->x - a2->x; 
	double e = a3->y - a2->y; 
	double f = a3->z - a2->z;

	double g = a4->x - a3->x; 
	double h = a4->y - a3->y; 
	double i = a4->z - a3->z;

	double ax = b*f - e*c; 
	double ay = c*d - f*a; 
	double az = a*e - b*d;

	double bx = h*f - e*i; 
	double by = i*d - f*g; 
	double bz = g*e - h*d;

        double xx = ay*bz - by*az;
        double yy = az*bx - bz*ax;
        double zz = ax*by - bx*ay;
 
        double cross = sqrt(xx*xx + yy*yy + zz*zz);
        double dot = ax*bx + ay*by + az*bz;

        double ang = fabs(atan2(cross, dot)) * RADIANS_TO_DEGREES;
  
	if ((ax*(e*bz-f*by) + ay*(f*bx-d*bz) +  az*(d*by-e*bx)) >=0.0e0)
		ang = -ang;

	return ang;
}

void Atom3d_IntToCart(Atom3d *a1)
{
	double dist;
	double rangl, rtors, sina, cosa, sint, cost;
	double u1x, u1y, u1z;
	double u2x, u2y, u2z;
	double u3x, u3y, u3z;
	double u4x, u4y, u4z;
	double sine, cosine, d;
	Atom3d *a2, *a3, *a4;

	if (a1->first_parent == NULL) {
		a1->x = a1->y = a1->z = 0.0e0;
		return ;
	}

	if (a1->second_parent == NULL) {
		a1->x = a1->fp_distance;
		a1->y = a1->z = 0.0e0;
		return;
	}

	if (a1->third_parent == NULL) {
		a1->z = 0.0e0;
		a1->x = ((Atom3d *)(a1->first_parent))->x  - \
			a1->fp_distance * cos(a1->sp_angle * 
					      DEGREES_TO_RADIANS);
		a1->y = a1->fp_distance * sin(a1->sp_angle * 
					      DEGREES_TO_RADIANS);
		return;
	}
		
	rangl = DEGREES_TO_RADIANS * (a1->sp_angle);
	rtors = DEGREES_TO_RADIANS * (a1->tp_torsion);
	dist = a1->fp_distance;

	sina = sin(rangl); cosa = -cos(rangl);
	sint = sina * sin(rtors); cost = sina * cos(rtors);

	a2 = (Atom3d *)a1->first_parent;
	a3 = (Atom3d *)a1->second_parent;
	a4 = (Atom3d *)a1->third_parent;

	u1x = a3->x - a4->x;
	u1y = a3->y - a4->y;
	u1z = a3->z - a4->z;
	d = 1.0e0 / sqrt(u1x*u1x + u1y*u1y + u1z*u1z);
	u1x *= d;
	u1y *= d;
	u1z *= d;

	u2x = a2->x - a3->x;
	u2y = a2->y - a3->y;
	u2z = a2->z - a3->z;
	d = 1.0e0 / sqrt(u2x*u2x + u2y*u2y + u2z*u2z);
	u2x *= d;
	u2y *= d;
	u2z *= d;

	cosine = u1x*u2x + u1y*u2y + u1z*u2z;

	if (fabs(cosine) < 1.0)
		sine = 1/sqrt(1.0 - cosine*cosine);
	else
		sine = 1/sqrt(cosine*cosine - 1.0);
            
	u3x = sine * (u1y*u2z - u1z*u2y);
	u3y = sine * (u1z*u2x - u1x*u2z);
	u3z = sine * (u1x*u2y - u1y*u2x);

	u4x = cost * (u3y*u2z - u3z*u2y);
	u4y = cost * (u3z*u2x - u3x*u2z);
	u4z = cost * (u3x*u2y - u3y*u2x);

	a1->x = a2->x + dist*(u2x*cosa + u4x + u3x*sint);
        a1->y = a2->y + dist*(u2y*cosa + u4y + u3y*sint);
        a1->z = a2->z + dist*(u2z*cosa + u4z + u3z*sint);

}

void Atom3d_ZmatrixToCart(int n, PyObject *atoms, int start, int end)
{
	int i;

        if (start < 0) start = 0;
	if (end < 0) end = n;

	for(i=start; i<end; i++) {
	        PyObject *tmp = PyList_GET_ITEM(atoms, i);
		Atom3d_IntToCart((Atom3d *)tmp);
	}
}

void Atom3d_GeoCenter(double *x, double *y, double *z, int n, PyObject *atoms)
{
	int i;
	double xc=0.0e0, yc=0.0e0, zc=0.0e0;

	for (i=0; i<n; i++) {
		PyObject *tmp = PyList_GET_ITEM(atoms, i);
		Atom3d *atom = (Atom3d *)tmp;
		xc += atom->x;
		yc += atom->y;
		zc += atom->z;
	}

	*x = xc/n;
	*y = yc/n;
	*z = zc/n;
}

double Atom3d_RadiusOfGyration(int n, PyObject *atoms)
{
	double xc, yc, zc, rg=0.0e0;
	int i;
	
	Atom3d_GeoCenter(&xc, &yc, &zc, n, atoms);
	for(i=0; i<n; i++) {
		PyObject *tmp = PyList_GET_ITEM(atoms, i);
		Atom3d *atom = (Atom3d *)tmp;
		double dx = atom->x - xc;
		double dy = atom->y - yc;
		double dz = atom->z - zc;
		rg += (dx*dx + dy*dy + dz*dz);
	}

	rg /= n;
	return sqrt(rg);
}
		
void Atom3d_RotateX(Atom3d *atom, double theta)
{
	double y, z;
	double trad = theta * DEGREES_TO_RADIANS;
	double sint = sin(trad);
	double cost = cos(trad);

	y = atom->y*cost - atom->z*sint;
	z = atom->y*sint + atom->z*cost;
	atom->y = y;
	atom->z = z;
}

void Atom3d_RotateY(Atom3d *atom, double theta)
{
	double x, z;
	double trad = theta * DEGREES_TO_RADIANS;
	double sint = sin(trad);
	double cost = cos(trad);

	x = atom->x*cost + atom->z*sint;
	z = atom->z*cost + atom->x*sint;
	atom->z = z;
	atom->x = x;
}

void Atom3d_RotateZ(Atom3d *atom, double theta)
{
	double x, y;
	double trad = theta * DEGREES_TO_RADIANS;
	double sint = sin(trad);
	double cost = cos(trad);

	x = atom->x*cost - atom->y*sint;
	y = atom->y*cost + atom->x*sint;
	atom->y = y;
	atom->x = x;
}

void Atom3d_BondRotate(Atom3d *batom1, Atom3d *batom2, PyObject *atoms, 
		       double theta)
{
	int n, i;

	double trad = theta * DEGREES_TO_RADIANS;
	double sinth = sin(trad);
	double costh = cos(trad);
	double dist = Atom3d_Distance(batom1, batom2);
	double ox = batom1->x;
	double oy = batom1->y;
	double oz = batom1->z;

	/* direction cosines of the bond axis */

	double lambda = (batom2->x - ox) / dist;
	double mu = (batom2->y - oy) / dist;
	double nu = (batom2->z - oz) / dist;

	/* construct the rotation matrix */

	double zip = 1.0e0 - costh;
	double rot00=costh+lambda*lambda*zip;
	double rot01=lambda*mu*zip-nu*sinth;
	double rot02=lambda*nu*zip+mu*sinth;
	double rot10=lambda*mu*zip+nu*sinth;
	double rot11=costh+mu*mu*zip;
	double rot12=mu*nu*zip-lambda*sinth;
	double rot20=lambda*nu*zip-mu*sinth;
	double rot21=mu*nu*zip+lambda*sinth;
	double rot22=costh+nu*nu*zip;

	/* apply rotation */

	n = PyList_GET_SIZE(atoms);

	for (i=0; i<n; i++) {
		Atom3d *atom = (Atom3d *)PyList_GET_ITEM(atoms, i);
		double dx = atom->x - ox;
		double dy = atom->y - oy;
		double dz = atom->z - oz;
		atom->x = ox + dx*rot00 + dy*rot01 + dz*rot02;
		atom->y = oy + dx*rot10 + dy*rot11 + dz*rot12;
		atom->z = oz + dx*rot20 + dy*rot21 + dz*rot22;
	}
}
	
static char at3d_create_doc[] = 
"Atom3d():\n\
	Create a new Atom3d object.\n\
";
static PyObject * at3d_create(PyObject *self, PyObject *args)
{
	Atom3d *atom;
	if (!PyArg_ParseTuple(args, ""))
		return NULL;
	atom = Atom3d_Create();
	if (atom == NULL)
		return NULL;
	else
		return (PyObject *)atom;
}

static char at3d_angle_doc[] =
"angle(a, b, c):\n\
	Compute the angle in degrees between atoms a, b and c:\n\
";
static PyObject *at3d_angle(PyObject *self, PyObject *args)
{
	PyObject *a, *b, *c;
	double ang;

	if (!(PyTuple_GET_SIZE(args) == 3)) {
		ErrorSet("angle(a, b, c) - requires 3 arguments");
		return NULL;
	}

	a = PyTuple_GET_ITEM(args, 0);
	if (!Atom3d_Check(a)) {
		ErrorSet("angle(a, b, c) - parameter a not Atom3d type");
		return NULL;
	}

	b = PyTuple_GET_ITEM(args, 1);
	if (!Atom3d_Check(b)) {
		ErrorSet("angle(a, b, c) - parameter b not Atom3d type");
		return NULL;
	}

	c = PyTuple_GET_ITEM(args, 2);
	if (!Atom3d_Check(c)) {
		ErrorSet("angle(a, b, c) - parameter c not Atom3d type");
		return NULL;
	}

	ang = Atom3d_Angle((Atom3d *)a, (Atom3d *)b, (Atom3d *)c);

	return PyFloat_FromDouble(ang);
}

static char at3d_torsion_doc[] =
"torsion(a, b, c, d):\n\
	Compute the angle in degrees between the normals of the planes\n\
	defined by atoms a,b and c and atoms b, c and d\n\
";
static PyObject *at3d_torsion(PyObject *self, PyObject *args)
{
	PyObject *a, *b, *c, *d;
	double ang;

	if (!(PyTuple_GET_SIZE(args) == 4)) {
		ErrorSet("torsion(a, b, c, d) - requires 4 arguments");
		return NULL;
	}

	a = PyTuple_GET_ITEM(args, 0);
	if (!Atom3d_Check(a)) {
		ErrorSet("torsion(a, b, c, d) - parameter a not Atom3d type");
		return NULL;
	}

	b = PyTuple_GET_ITEM(args, 1);
	if (!Atom3d_Check(b)) {
		ErrorSet("torsion(a, b, c, d) - parameter b not Atom3d type");
		return NULL;
	}

	c = PyTuple_GET_ITEM(args, 2);
	if (!Atom3d_Check(c)) {
		ErrorSet("torsion(a, b, c, d) - parameter c not Atom3d type");
		return NULL;
	}

	d = PyTuple_GET_ITEM(args, 3);
	if (!Atom3d_Check(d)) {
		ErrorSet("torsion(a, b, c, d) - parameter d not Atom3d type");
		return NULL;
	}

	ang = Atom3d_Torsion((Atom3d *)a, (Atom3d *)b, (Atom3d *)c,
			     (Atom3d *)d);

	return PyFloat_FromDouble(ang);
}
static char at3d_ztox_doc[] =
"ztox(atomArray, [start=0]):\n\
	Compute cartesian coordinates for an array of atoms, beginning at\n\
	index start.  The parameter atomArray must be a Numeric array of\n\
	atoms\n\
";
static PyObject * at3d_ztox(PyObject *self, PyObject *args)
{
	PyObject *atoms;
	int start=0;
        int end=-1;
	int n, nargs;

	nargs = PyTuple_GET_SIZE(args);

	if (nargs <1 || nargs >3) {
		ErrorSet("ztox(atom, [start], [end]) - requires at most 3 \
arguments");
		return NULL;
	}

	atoms = PyTuple_GET_ITEM(args, 0);
	if (!PyList_Check(atoms)) {
		ErrorSet("ztox(atom, [start], [end]) - parameter atoms "\
			 "must be a list of atoms");
		return NULL;
	}

	if (nargs > 1) {
		PyObject *tmp = PyTuple_GET_ITEM(args, 1);
		if (PyInt_Check(tmp))
			start = (int )PyInt_AS_LONG(tmp);
		else {
			ErrorSet("ztox(atom, start, [end]) - parameter "\
				 "start must  be an integer");
			return NULL;
		}
	}

	n = PyList_GET_SIZE(atoms);

	if (nargs == 3) {
		PyObject *tmp = PyTuple_GET_ITEM(args, 2);
		if (PyInt_Check(tmp))
			end = (int )PyInt_AS_LONG(tmp);
		else {
			ErrorSet("ztox(atom, start, end) - parameter end "\
				 "must be an integer");
			return NULL;
		}
	}

	Atom3d_ZmatrixToCart(n, atoms, start, end);
	Py_INCREF(Py_None);
	return Py_None;
}
	
static char at3d_radOfGyr_doc[] =
"radius_of_gyration(atomList):\n\
	Compute the radius of gyration of a list of atoms\n\
";

static PyObject *at3d_radOfGyr(PyObject *self, PyObject *args)
{
	PyObject *atoms;
	int n;
	
	if (!(PyTuple_GET_SIZE(args) == 1)) {
		ErrorSet("radOfGyr(atoms) - requires a list of atoms " \
			 "as argument");
		return NULL;
	}

	atoms = PyTuple_GET_ITEM(args, 0);
	if (!PyList_Check(atoms)) {
		ErrorSet("radOfGyr(atoms) - requires a list of atoms "\
			 "as argument");
		return NULL;
	}

	n = PyList_GET_SIZE(atoms);
	return PyFloat_FromDouble(Atom3d_RadiusOfGyration(n, atoms));
}

static char at3d_geoCen_doc[] =
"geometric_center(atomList):\n\
	Returns a tuple of x,y,z coordinates representing the geometric center\n\
	of a list of atoms\n\
";

static PyObject *at3d_geoCen(PyObject *self, PyObject *args)
{
	PyObject *atomList;
	int n;
	double x, y, z;
	PyObject *result;
	
	if (!(PyTuple_GET_SIZE(args) == 1)) {
		ErrorSet("geometricCenter(atoms) - requires a list of atoms as \
argument");
		return NULL;
	}

	atomList = PyTuple_GET_ITEM(args, 0);
	if (!PyList_Check(atomList)) {
		ErrorSet("geometricCenter(atoms) - requires a list of atoms as \
argument");
		return NULL;
	}

	n = PyList_GET_SIZE(atomList);
	Atom3d_GeoCenter(&x, &y, &z, n, atomList);
	result = PyTuple_New(3);
	if (result == NULL)
		return NULL;
	PyTuple_SetItem(result, 0, PyFloat_FromDouble(x));
	PyTuple_SetItem(result, 1, PyFloat_FromDouble(y));
	PyTuple_SetItem(result, 2, PyFloat_FromDouble(z));
        return result;
}

static PyObject *at3d_cic(PyObject *self, PyObject *args)
{
	int start=0;
	int end=-1;
	PyObject *atomList;
	int n, i;

	if (!PyArg_ParseTuple(args, "O|ii", &atomList, &start, &end))
		return NULL;
	if (!PyList_Check(atomList)) {
		ErrorSet("commit_internal_coords - parameter atomList must be \
python list");
		return NULL;
	}

	n = PyList_GET_SIZE(atomList);
	if (end == -1) end = n;

	for (i=start; i<end; i++) {
		Atom3d *atom = (Atom3d *)PyList_GET_ITEM(atomList, i);
		atom->fp_distance_tmp = atom->fp_distance;
		atom->sp_angle_tmp = atom->sp_angle;
		atom->tp_torsion_tmp = atom->tp_torsion;
	}

	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *at3d_ric(PyObject *self, PyObject *args)
{
	int start=0;
	int end=-1;
	PyObject *atomList;
	int n, i;

	if (!PyArg_ParseTuple(args, "O|ii", &atomList, &start, &end))
		return NULL;
	if (!PyList_Check(atomList)) {
		ErrorSet("restore_internal_coords - parameter atomList must be \
python list");
		return NULL;
	}

	n = PyList_GET_SIZE(atomList);
	if (end == -1) end = n;

	for (i=start; i<end; i++) {
		Atom3d *atom = (Atom3d *)PyList_GET_ITEM(atomList, i);
		atom->fp_distance = atom->fp_distance_tmp;
		atom->sp_angle = atom->sp_angle_tmp;
		atom->tp_torsion = atom->tp_torsion_tmp;
	}

	Py_INCREF(Py_None);
	return Py_None;
}
		
static char at3d_bond_rotate_doc[] =
"bond_rotate(b1, b2, atoms, theta):\n\
	rotate atoms by thetat degrees about the bond defined by atoms\n\
        b1 and b2.  Parameter atoms must be a python list of atoms\n\
";

static PyObject *at3d_bond_rotate(PyObject *self, PyObject *args)
{
	PyObject *b1, *b2, *atoms;
	double theta;

	if (!PyArg_ParseTuple(args, "OOOd", &b1, &b2, &atoms, &theta))
		return NULL;
	if (!Atom3d_Check(b1)){
		ErrorSet("bond_rotate(b1, b2, atoms, theta) - b1 must be Atom3d type");
		return NULL;
	}

	if (!Atom3d_Check(b2)){
		ErrorSet("bond_rotate(b1, b2, atoms, theta) - b2 must be Atom3d type");
		return NULL;
	}

	if (!PyList_Check(atoms)) {
		ErrorSet("bond_rotate(b1, b2, atoms, theta) - atoms must be Python List");
		return NULL;
	}
		
	Atom3d_BondRotate((Atom3d *)b1, (Atom3d *)b2, atoms, theta);

	Py_INCREF(Py_None);
	return Py_None;
}

static struct PyMethodDef atom3d_methods[] = {
	{"Atom3d",		(PyCFunction)at3d_create,	METH_VARARGS,
	at3d_create_doc},
	{"angle",		(PyCFunction)at3d_angle,	METH_VARARGS,
	at3d_angle_doc},
	{"torsion",		(PyCFunction)at3d_torsion,	METH_VARARGS,
	at3d_torsion_doc},
	{"ztox",		(PyCFunction)at3d_ztox,		METH_VARARGS,
	at3d_ztox_doc},
	{"radius_of_gyration",	(PyCFunction)at3d_radOfGyr,	METH_VARARGS,
	at3d_radOfGyr_doc},
	{"geometric_center",	(PyCFunction)at3d_geoCen,	METH_VARARGS,
	at3d_geoCen_doc},
	{"commit_internal_coords", (PyCFunction)at3d_cic,	METH_VARARGS,
	NULL},
	{"restore_internal_coords", (PyCFunction)at3d_ric,	METH_VARARGS,
	NULL},
	{"bond_rotate",		(PyCFunction)at3d_bond_rotate,	METH_VARARGS,
	at3d_bond_rotate_doc},
	{NULL, NULL, 0, NULL}
};

static char at3d_distance_doc[] =
"a.distance(b):\n\
	Compute the euclidean distance between atoms a and b\n\
";

static PyObject *at3d_distance(Atom3d *self, PyObject *args)
{
	PyObject *tmp;
	Atom3d *atom;
	double dx, dy, dz;

	if (!(PyTuple_GET_SIZE(args)==1)) {
		ErrorSet("Atom3d.distance(a): requires one argument");
		return NULL;
	}
        tmp = PyTuple_GET_ITEM(args, 0);
	if (!Atom3d_Check(tmp)) {
		ErrorSet("Atom3d.distance(a): a must be Atom3d type");
		return NULL;
	}
	atom = (Atom3d *)tmp;
	dx = self->x - atom->x;
	dy = self->y - atom->y;
	dz = self->z - atom->z;
	return PyFloat_FromDouble(sqrt(dx*dx + dy*dy + dz*dz));
}

static char at3d_sqdistance_doc[] =
"a.sqdistance(b):\n\
	Compute the square of the euclidean distance between atoms a and b\n\
";

static PyObject *at3d_sqdistance(Atom3d *self, PyObject *args)
{
	PyObject *tmp;
	Atom3d *atom;
	double dx, dy, dz;

	if (!(PyTuple_GET_SIZE(args)==1)) {
		ErrorSet("Atom3d.sqdistance(a): requires one argument");
		return NULL;
	}
	tmp = PyTuple_GET_ITEM(args, 0);
	if (!Atom3d_Check(tmp)) {
		ErrorSet("Atom3d.sqdistance(a): a must be Atom3d type");
		return NULL;
	}
	atom = (Atom3d *)tmp;
	dx = self->x - atom->x;
	dy = self->y - atom->y;
	dz = self->z - atom->z;
	return PyFloat_FromDouble(dx*dx + dy*dy + dz*dz);
}

static char at3d_close_doc[] =
"a.close(b, [dist]):\n\
	Check if the distance between atoms a and b is <= dist.\n\
	If parameter dist is omitted the sum of the atom radii is \n\
	is used.\n\
";
static PyObject *at3d_close(Atom3d *self, PyObject *args)
{
	PyObject *tmp;
	Atom3d *atom;
	double tol, dx, dy, dz;

	int nargs = PyTuple_GET_SIZE(args);

	if (nargs < 1 || nargs > 2) {
		ErrorSet("Atom3d.close(a, [dist]): requires 1 or 2 arguments");
		return NULL;
	}
	tmp = PyTuple_GET_ITEM(args, 0);
	if (!Atom3d_Check(tmp)) {
		ErrorSet("Atom3d.close(a, [dist]): a must be Atom3d type");
		return NULL;
	}
	atom = (Atom3d *)tmp;

	if (nargs == 2) {
		tmp = PyTuple_GET_ITEM(args, 1);
		if (PyFloat_Check(tmp))
			tol = PyFloat_AS_DOUBLE(tmp);
		else if (PyInt_Check(tmp))
			tol = (double )PyInt_AS_LONG(tmp);
		else {
			ErrorSet("Atom3d.close(a, dist): dist must be float type");
			return NULL;
		}
	}
	else {
		tol = self->radius + atom->radius;
		if (tol <= 1.0e-6) return PyInt_FromLong(0L);
	}

	dx = fabs(self->x - atom->x);
	if (dx >= tol)
		return PyInt_FromLong(0L);
	dy = fabs(self->y - atom->y);
	if (dy >= tol)
		return PyInt_FromLong(0L);
	dz = fabs(self->z - atom->z);
	if (dz >= tol)
		return PyInt_FromLong(0L);
	if ((dx*dx + dy*dy + dz*dz) > tol*tol)
		return PyInt_FromLong(0L);
	return PyInt_FromLong(1L);
}

static char at3d_IntToCart_doc[] =
"a.intToCart():\n\
	Compute atom's cartesian coordinates from internal coordinate\n\
	description.  Prior to calling this function the atom's first,\n\
	second and third parent's and the associated distance, angle and\n\
	torsion must have been set.\n\
";

static PyObject *at3d_IntToCart(Atom3d *self, PyObject *args)
{
	Atom3d_IntToCart(self);
	Py_INCREF(Py_None);
	return Py_None;
}

static char at3d_coords_doc[] =
"a.coords():\n\
	returns the atoms cartesian coordinates as a python list\n\
";

static PyObject *at3d_coords(Atom3d *self, PyObject *args)
{
	PyObject *result;

	if (!PyArg_ParseTuple(args, "")) return NULL;
	result = PyList_New(3);
	if (result == NULL) return NULL;
	PyList_SetItem(result, 0, PyFloat_FromDouble(self->x));
	PyList_SetItem(result, 1, PyFloat_FromDouble(self->y));
	PyList_SetItem(result, 2, PyFloat_FromDouble(self->z));
	return result;
}

static PyObject *at3d_coic(Atom3d *self, PyObject *args)
{
	if (!PyArg_ParseTuple(args, "")) return NULL;
	self->fp_distance_tmp = self->fp_distance;
	self->sp_angle_tmp = self->sp_angle;
	self->tp_torsion_tmp = self->tp_torsion;
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *at3d_reic(Atom3d *self, PyObject *args)
{
	if (!PyArg_ParseTuple(args, "")) return NULL;
	self->fp_distance = self->fp_distance_tmp;
	self->sp_angle = self->sp_angle_tmp;
	self->tp_torsion = self->tp_torsion_tmp;
	Py_INCREF(Py_None);
	return Py_None;
}

static char at3d_rotatex_doc[] =
"a.rotatex(theta):\n\
	Rotate atom by theta degrees about X axis.\n\
";


static PyObject *at3d_rotatex(Atom3d *self, PyObject *args)
{
	PyObject *ptheta;
	double theta;

	if (!(PyTuple_GET_SIZE(args) == 1)) {
		ErrorSet("Atom3d.rotatex(theta) - requires one argument");
		return NULL;
	}

	ptheta = PyTuple_GET_ITEM(args, 0);
	if (PyFloat_Check(ptheta))
		theta = PyFloat_AS_DOUBLE(ptheta);
	else if (PyInt_Check(ptheta))
		theta = (double )PyInt_AS_LONG(ptheta);
	else {
		ErrorSet("Atom3d.rotatex(theta) - theta must be type float");
		return NULL;
	}

	Atom3d_RotateX(self, theta);
	Py_INCREF(Py_None);
	return Py_None;
}

static char at3d_rotatey_doc[] =
"a.rotatey(theta):\n\
	Rotate atom by theta degrees about Y axis.\n\
";
static PyObject *at3d_rotatey(Atom3d *self, PyObject *args)
{
	PyObject *ptheta;
	double theta;

	if (!(PyTuple_GET_SIZE(args) == 1)) {
		ErrorSet("Atom3d.rotatey(theta) - requires one argument");
		return NULL;
	}

	ptheta = PyTuple_GET_ITEM(args, 0);
	if (PyFloat_Check(ptheta))
		theta = PyFloat_AS_DOUBLE(ptheta);
	else if (PyInt_Check(ptheta))
		theta = (double )PyInt_AS_LONG(ptheta);
	else {
		ErrorSet("Atom3d.rotatey(theta) - theta must be type float");
		return NULL;
	}

	Atom3d_RotateY(self, theta);
	Py_INCREF(Py_None);
	return Py_None;
}

static char at3d_rotatez_doc[] =
"a.rotatez(theta):\n\
	Rotate atom by theta degrees about Z axis.\n\
";
static PyObject *at3d_rotatez(Atom3d *self, PyObject *args)
{
	PyObject *ptheta;
	double theta;

	if (!(PyTuple_GET_SIZE(args) == 1)) {
		ErrorSet("Atom3d.rotatez(theta) - requires one argument");
		return NULL;
	}

	ptheta = PyTuple_GET_ITEM(args, 0);
	if (PyFloat_Check(ptheta))
		theta = PyFloat_AS_DOUBLE(ptheta);
	else if (PyInt_Check(ptheta))
		theta = (double )PyInt_AS_LONG(ptheta);
	else {
		ErrorSet("Atom3d.rotatez(theta) - theta must be type float");
		return NULL;
	}

	Atom3d_RotateZ(self, theta);
	Py_INCREF(Py_None);
	return Py_None;
}

static struct PyMethodDef at3d_methods[] = {
	{"distance",	(PyCFunction)at3d_distance,	METH_VARARGS,	at3d_distance_doc},
	{"sqdistance",	(PyCFunction)at3d_sqdistance,	METH_VARARGS,	at3d_sqdistance_doc},
	{"close",	(PyCFunction)at3d_close,	METH_VARARGS,	at3d_close_doc},
	{"int_to_cart",	(PyCFunction)at3d_IntToCart,	METH_VARARGS,	at3d_IntToCart_doc},
	{"coords",	(PyCFunction)at3d_coords,	METH_VARARGS,	at3d_coords_doc},
	{"commit_internal_coords", (PyCFunction)at3d_coic, METH_VARARGS, NULL},
	{"restore_internal_coords", (PyCFunction)at3d_reic, METH_VARARGS, NULL},
	{"rotatex",	(PyCFunction)at3d_rotatex,	METH_VARARGS,	at3d_rotatex_doc},
	{"rotatey",	(PyCFunction)at3d_rotatey,	METH_VARARGS,	at3d_rotatey_doc},
	{"rotatez",	(PyCFunction)at3d_rotatez,	METH_VARARGS,	at3d_rotatez_doc},
 
	{NULL, NULL, 0, NULL}
};

/* atom3d handling */
static void at3d_dealloc(Atom3d *self)
{
	Py_XDECREF(self->first_parent);
	Py_XDECREF(self->second_parent);
	Py_XDECREF(self->third_parent);
	PyMem_DEL(self);

}

static PyObject *at3d_getattr(Atom3d *self, char *name)
{
	PyObject *result = Py_FindMethod(at3d_methods, (PyObject *)self, name);
	if (result == NULL)
		PyErr_Clear();
	else
		return result;

	if (strcmp("bsep0", name) == 0)
		return PyInt_FromLong((long )(self->bsep0));
	else if (strcmp("bsep1", name) == 0)
		return PyInt_FromLong((long )(self->bsep1));
	else if (strcmp("resnum", name) == 0)
		return PyInt_FromLong((long )(self->resnum));
	else if (strcmp("first_parent", name) == 0) {
		if (self->first_parent == NULL) {
			Py_INCREF(Py_None);
			return Py_None;
		}
		else {
			Py_INCREF(self->first_parent);
			return self->first_parent;
		}
	}
	else if (strcmp("second_parent", name) == 0) {
		if (self->second_parent == NULL) {
			Py_INCREF(Py_None);
			return Py_None;
		}
		else {
			Py_INCREF(self->second_parent);
			return self->second_parent;
		}
	}
	else if (strcmp("third_parent", name) == 0) {
		if (self->third_parent == NULL) {
			Py_INCREF(Py_None);
			return Py_None;
		}
		else {
			Py_INCREF(self->third_parent);
			return self->third_parent;
		}
	}
	else if (strcmp("fp_distance", name) == 0)
		return PyFloat_FromDouble(self->fp_distance);
	else if (strcmp("sp_angle", name) == 0)
		return PyFloat_FromDouble(self->sp_angle);
	else if (strcmp("tp_torsion", name) == 0)
		return PyFloat_FromDouble(self->tp_torsion);
	else if (strcmp("x", name) == 0)
		return PyFloat_FromDouble(self->x);
	else if (strcmp("y", name) == 0)
		return PyFloat_FromDouble(self->y);
	else if (strcmp("z", name) == 0)
		return PyFloat_FromDouble(self->z);
	else if (strcmp("name", name) == 0)
		return PyString_FromString(self->name);
	else if (strcmp("radius", name) == 0)
		return PyFloat_FromDouble(self->radius);
	else if (strcmp("sasa", name) == 0)
		return PyFloat_FromDouble(self->sasa);
	else if (strcmp("sasarad", name) == 0)
		return PyFloat_FromDouble(self->sasarad);
	else if (strcmp("sasapar", name) == 0)
		return PyFloat_FromDouble(self->sasapar);
	else {
		PyErr_SetString(PyExc_AttributeError,name);
		return NULL;
	}

}

static int at3d_setattr(Atom3d *self, char *name, PyObject *value)
{
	char buf[255];
	char *floatError = "Expected a float for attribute %s";
	char *intError = "Expected an integer for attribute %s";
	char *strError = "Expected a string for attribute %s";
	char *atomError = "Expected Atom3d type for attribute %s";

	if (value == NULL) {
		PyErr_SetString(PyExc_AttributeError,
				"Object attributes cannot be deleted");
		return -1;
	}
	if (strcmp("first_parent", name) == 0) {
		if (!Atom3d_Check(value)) {
			sprintf(buf, atomError, "first_parent");
			PyErr_SetString(PyExc_TypeError, buf);
			return -1;
		}
		else if (((Atom3d *)value) == self) {
			PyErr_SetString(PyExc_ValueError, "atom and \
first_parent are identical");
			return -1;
		}
		else {
			Py_INCREF(value);
			self->first_parent = value; 
			return 0;
		}
	}
	else if (strcmp("second_parent", name) == 0) {
		if (!Atom3d_Check(value)) {
			sprintf(buf, atomError, "second_parent");
			PyErr_SetString(PyExc_TypeError, buf);
			return -1;
		}
		else if (((Atom3d *)value) == self) {
			PyErr_SetString(PyExc_ValueError, "atom and \
second_parent are identical");
			return -1;
		}
		else if (value == self->first_parent) {
			PyErr_SetString(PyExc_ValueError, "atom's first and \
second parents are identical");
			return -1;
		}
		else {
			Py_INCREF(value);
			self->second_parent = value;
			return 0;
		}
	}
	else if (strcmp("third_parent", name) == 0) {
		if (!Atom3d_Check(value)) {
			sprintf(buf, atomError, "third_parent");
			PyErr_SetString(PyExc_TypeError, buf);
			return -1;
		}
		else if (((Atom3d *)value) == self) {
			PyErr_SetString(PyExc_ValueError, "atom and \
third_parent are identical");
			return -1;
		}
		else if (value == self->first_parent) {
			PyErr_SetString(PyExc_ValueError, "atom's first and \
third parents are identical");
			return -1;
		}
		else if (value == self->second_parent) {
			PyErr_SetString(PyExc_ValueError, "atom's second and \
third parents are identical");
			return -1;
		}
		else {
			Py_INCREF(value);
			self->third_parent = value;
			return 0;
		}
	}
	else if (strcmp("fp_distance", name) == 0) {
		if (PyFloat_Check(value)) {
			self->fp_distance = PyFloat_AS_DOUBLE(value);
			return 0;
		}
		if (PyInt_Check(value)) {
			double tmp = (double )PyInt_AS_LONG(value);
			self->fp_distance = tmp;
			return 0;
		}
		sprintf(buf, floatError, "fp_distance");
		PyErr_SetString(PyExc_TypeError, buf);
		return -1;
	}
	else if (strcmp("sp_angle", name) == 0) {
		if (PyFloat_Check(value)) {
			self->sp_angle = PyFloat_AS_DOUBLE(value);
			return 0;
		}
		if (PyInt_Check(value)) {
			double tmp = (double )PyInt_AS_LONG(value);
			self->sp_angle = tmp;
			return 0;
		}
		sprintf(buf, floatError, "sp_angle");
		PyErr_SetString(PyExc_TypeError, buf);
		return -1;
	}
	else if (strcmp("tp_torsion", name) == 0) {
		if (PyFloat_Check(value)) {
			self->tp_torsion = PyFloat_AS_DOUBLE(value);
			return 0;
		}
		if (PyInt_Check(value)) {
			double tmp = (double )PyInt_AS_LONG(value);
			self->tp_torsion = tmp;
			return 0;
		}
		sprintf(buf, floatError, "tp_torsion");
		PyErr_SetString(PyExc_TypeError, buf);
		return -1;
	}
	else if (strcmp("x", name) == 0) {
		if (PyFloat_Check(value)) {
			self->x = PyFloat_AS_DOUBLE(value);
			return 0;
		}
		if (PyInt_Check(value)) {
			double tmp = (double )PyInt_AS_LONG(value);
			self->x = tmp;
			return 0;
		}
		sprintf(buf, floatError, "x");
		PyErr_SetString(PyExc_TypeError, buf);
		return -1;
	}
	else if (strcmp("y", name) == 0) {
		if (PyFloat_Check(value)) {
			self->y = PyFloat_AS_DOUBLE(value);
			return 0;
		}
		if (PyInt_Check(value)) {
			double tmp = (double )PyInt_AS_LONG(value);
			self->y = tmp;
			return 0;
		}
		sprintf(buf, floatError, "y");
		PyErr_SetString(PyExc_TypeError, buf);
		return -1;
	}

	else if (strcmp("z", name) == 0) {
		if (PyFloat_Check(value)) {
			self->z = PyFloat_AS_DOUBLE(value);
			return 0;
		}
		if (PyInt_Check(value)) {
			double tmp = (double )PyInt_AS_LONG(value);
			self->z = tmp;
			return 0;
		}
		sprintf(buf, floatError, "z");
		PyErr_SetString(PyExc_TypeError, buf);
		return -1;
	}

	else if (strcmp("name", name) == 0) {
		if (!PyString_Check(value)) {
			sprintf(buf, strError, "name");
			PyErr_SetString(PyExc_TypeError, buf);
			return -1;
		}
		else {
			if (PyString_Size(value) > 4) {
				PyErr_SetString(PyExc_ValueError,
						"name must be length 4 or less"
						);
				return -1;
			}
			strncpy(self->name, "    \0", 5);
			strncpy(self->name, PyString_AsString(value),
				PyString_Size(value));
			return 0;
		}
	}
	else if (strcmp("radius", name) == 0) {
		if (PyFloat_Check(value)) {
			self->radius = PyFloat_AS_DOUBLE(value);
			return 0;
		}
		if (PyInt_Check(value)) {
			double tmp = (double )PyInt_AS_LONG(value);
			self->radius = tmp;
			return 0;
		}
		sprintf(buf, floatError, "radius");
		PyErr_SetString(PyExc_TypeError, buf);
		return -1;
	}
	else if (strcmp("bsep0", name) == 0) {
		if (PyInt_Check(value)) {
			self->bsep0 = (int )PyInt_AS_LONG(value);
			return 0;
		}
		sprintf(buf, intError, "bsep0");
		PyErr_SetString(PyExc_TypeError, buf);
		return -1;
	}
	else if (strcmp("bsep1", name) == 0) {
		if (PyInt_Check(value)) {
			self->bsep1 = (int )PyInt_AS_LONG(value);
			return 0;
		}
		sprintf(buf, intError, "bsep1");
		PyErr_SetString(PyExc_TypeError, buf);
		return -1;
	}
	else if (strcmp("resnum", name) == 0) {
		if (PyInt_Check(value)) {
			self->resnum = (int )PyInt_AS_LONG(value);
			return 0;
		}
		sprintf(buf, intError, "resnum");
		PyErr_SetString(PyExc_TypeError, buf);
		return -1;
	}
	else if (strcmp("sasa", name) == 0) {
		if (PyFloat_Check(value)) {
			self->sasa = PyFloat_AS_DOUBLE(value);
			return 0;
		}
		sprintf(buf, floatError, "sasa");
		return -1;
	}
	else if (strcmp("sasarad", name) == 0) {
		if (PyFloat_Check(value)) {
			self->sasarad = PyFloat_AS_DOUBLE(value);
			return 0;
		}
		sprintf(buf, floatError, "sasarad");
		return -1;
	}
	else if (strcmp("sasapar", name) == 0) {
		if (PyFloat_Check(value)) {
			self->sasapar = PyFloat_AS_DOUBLE(value);
			return 0;
		}
		sprintf(buf, floatError, "sasapar");
		return -1;
	}
	else {
		PyErr_SetString(PyExc_AttributeError,name);
		return -1;
	}

}

static int at3d_hash(Atom3d *self, PyObject *args)
{
	char buf[100];
	register int len, n;
	register unsigned char *p;
	register long x;

	len = sprintf(buf, "%d_%s", self->resnum, self->name);
	n = len;
        p = (unsigned char *) buf;
	x = *p << 7;
	while (--len >= 0)
		x = (1000003*x) ^ *p++;
	x ^= n;
	if (x == -1)
		x = -2;
	return x;
}

statichere PyTypeObject Atom3dType = {
	PyObject_HEAD_INIT(NULL)
	0,
	"Atom3d",
	sizeof(Atom3d),
	0,
	(destructor)at3d_dealloc,
	0,
	(getattrfunc)at3d_getattr,
	(setattrfunc)at3d_setattr,
	0,
	0,
	0,
	0,
	0,
	(hashfunc)at3d_hash,
	0,
	0,
	0, 0, 0, 0,
        NULL
};

DL_EXPORT(void)
 initAtom3dc(void)
{
    PyObject *m, *d;
    static void *ATOM3D_API[ATOM3D_API_POINTERS];

    Atom3dType.ob_type = &PyType_Type;
    m = Py_InitModule4("Atom3dc", atom3d_methods,
                       "",
                       (PyObject *)NULL, PYTHON_API_VERSION);

    d = PyModule_GetDict(m);
    ErrorObject = PyString_FromString("Atom3dc.error");
    PyDict_SetItemString(d, "error", ErrorObject);

    ATOM3D_API[Atom3d_Type_NUM] = (void *)&Atom3dType;
    ATOM3D_API[Atom3d_Create_NUM] = (void *)&Atom3d_Create;
    ATOM3D_API[Atom3d_SqDistance_NUM] = (void *)&Atom3d_SqDistance;
    ATOM3D_API[Atom3d_Distance_NUM] = (void *)&Atom3d_Distance;
    ATOM3D_API[Atom3d_Angle_NUM] = (void *)&Atom3d_Angle;
    ATOM3D_API[Atom3d_Torsion_NUM] = (void *)&Atom3d_Torsion;
    ATOM3D_API[Atom3d_GeoCenter_NUM] = (void *)&Atom3d_GeoCenter;
    ATOM3D_API[Atom3d_IntToCart_NUM] = (void *)&Atom3d_IntToCart;
    ATOM3D_API[Atom3d_ZmatrixToCart_NUM] = (void *)&Atom3d_ZmatrixToCart;
    ATOM3D_API[Atom3d_RadiusOfGyration_NUM] = (void *)&Atom3d_RadiusOfGyration;
    ATOM3D_API[Atom3d_RotateX_NUM] = (void *)&Atom3d_RotateX;
    ATOM3D_API[Atom3d_RotateY_NUM] = (void *)&Atom3d_RotateY;
    ATOM3D_API[Atom3d_RotateZ_NUM] = (void *)&Atom3d_RotateZ;
    ATOM3D_API[Atom3d_BondRotate_NUM] = (void *)&Atom3d_BondRotate;

    PyDict_SetItemString(d, "_C_API",
                         PyCObject_FromVoidPtr((void *)ATOM3D_API, NULL));
    if (PyErr_Occurred())
        Py_FatalError("can't initialize module Atom3d");
}
