"""Module implementing the scoring functions in LINUS

Functions in this module

    o *torsion_score* - compute torsion score of a molecule

    o *print_torsion_score* - print details of torsion score to file

    o *contact_score* - compute hydrophobic contact energy of a molecule

    o *print_contact_score* - print details of contact energy to file

    o *hbond_score* - compute hydrogen bond energy of a molecule

    o *print_hbond_score* - print details of hydrogen bond energy to
    file
"""

import Atom3d
import sys
import operator
import math

def torsion_score(mol, tp=1.0):
    """Torsion score of a molecule

    Arguments

        o *mol* - instance of LinusMol - molecule whose torsion
        score is to be computed

        o *tp* - float - value of torsion penalty

    Result

        The torsion score as a floating point number
    """
    
    rnames = mol.residue_names
    pos = 0
    etors = 0.0

    for atom in mol.phi_atoms:
        if 0.0 < atom.tp_torsion < 150.0:
            if rnames[pos] is 'GLY':
                etors = etors - tp
            elif rnames[pos] is 'ASN':
                etors = etors
            else:
                etors = etors + tp
        pos = pos + 1
    return etors

def print_torsion_score(mol, tp=1.0, file=sys.stdout):
    """List torsion score of a molecule to a file

    Arguments

        o *mol* - instance of LinusMol - molecule whose torsion
        score is to be computed

        o *tp* - float - value of torsion penalty

        o *file* - string or file object - either the name of a
        file or a file that has been previously opened for writing

    Result

        The torsion score as a floating point number
    """
    if type(file) is type(''):
        ff = open(file, 'wb')
    else:
        ff = file
    ff.write('\n------ Torsion Penalty Listing -----\n')
    rnames = mol.residue_names
    pos = 0
    etors = 0.0

    for atom in mol.phi_atoms:
        phi = atom.tp_torsion
        if 0.0 < phi < 150.0:
            if  rnames[pos] is ('GLY'):
                etors = etors - tp
            else:
                etors = etors + tp
            ff.write('%d %s %.2f %.2f\n' % (pos+1, rnames[pos], phi, tp))
        pos = pos + 1
    ff.write('\nTotal torsion penalty = %.2f\n' % etors)
    if ff <> file: ff.close()
    return etors

        
def compaction_score(mol, cp=1.0, rg=Atom3d.radius_of_gyration,
                     exp=math.exp):
    return rg(mol.atoms)

def print_compaction_score(mol, cp=1.0, file=sys.stdout):
    if type(file) is type(''):
        ff = open(file, 'wb')
    else:
        ff = file
    ff.write('\n------ Compaction Score -----\n')
    rg = Atom3d.radius_of_gyration(mol.atoms)
    ff.write('Radius of Gyration = %g\n' % rg)
    ff.write('\nCompaction Score = %g\n' % (math.exp(-rg*rg),))
    if ff <> file: ff.close()
    return rg*cp    

def contact_score(clist, cprobe):
    """Compute contact score for a molecule

    Arguments

        o *clist* - object - the object returned by the *make_contact_list*
        function in the *construct* module

        o *cprobe* - float - the distance over which the contact energy
        is to be scaled to zero.

    Result

        The contact energy as a floating point number
    """
    
    eborpc = 1.0/cprobe
    econ = 0.0
    for c1, c2, dmin, ene in clist:
        dmax = dmin + cprobe
        if c1.close(c2, dmax):
            d12 = c1.distance(c2)
##            if not c1.first_parent.close(c2.first_parent, d12):
            if d12 < dmin:
                econ = econ + ene
            else:
                econ = econ + ene*(dmax-d12)*eborpc
    return -econ

def print_contact_score(mol, clist, cprobe, file):
    """Print contact energy details to a file

    Arguments

        o *mol* - instance of LinusMol - molecule whose contact score
        is to be computed and listed to file

        o *clist* - object - the object returned by the *make_contact_list*
        function in the *construct* module

        o *cprobe* - float - the distance over which the contact energy
        is to be scaled to zero.

        o *file* - string or file object - name of a file or a file
        previously opened for writing.

    Result

        The contact energy of the molecule
    """

    if type(file) is type(''):
        ff = open(file, 'wb')
    else:
        ff = file

    rnames = mol.residue_names
    atoms = mol.atoms
    fai = mol.residue_first_atom_indices

    Write = ff.write
    Write('\n------ Contact Score Listing -----\n')
    eborpc = 1.0/cprobe
    econ = 0.0
    pos = 0
    for c1, c2, dmin, ene in clist:
        dmax = dmin + cprobe
        if c1.close(c2, dmax):
            d12 = c1.distance(c2)
##            if not c1.first_parent.close(c2.first_parent, d12):
            if d12 < dmin:
                etmp = ene
            else:
                etmp = ene*(dmax-d12)*eborpc

            Write('%3d %s %s %3d %s %s Distance = %.3f Score = %.3f\n' % 
                   (c1.resnum+1, rnames[c1.resnum], c1.name,
                    c2.resnum+1, rnames[c2.resnum], c2.name,
                    d12, etmp))
            econ = econ + etmp
        pos = pos + 1
        
    Write('\nTotal Contact Score = %g\n' % -econ)

    if ff <> file:
        ff.close()
    return -econ

def hbond_score(hblist, ABS=abs, ANGLE=Atom3d.angle,
                        TORSION=Atom3d.torsion):
    """Compute hydrogen bond energy of a molecule

    Arguments

        o *hblist* - object - the object returned by the function
        *make_hbond_list* in the *construct* module

    Result

        The hydrogen bond energy as a floating point number
    """    
    ehb = 0.0
    for dlist, alist in hblist:
        d1, da1, da2 = dlist
        Close = d1.close
        for acp, hbd, hbdmax, hbtors, hbene in alist:
            if Close(acp, hbdmax):
                if ABS(TORSION(acp, d1, da1, da2)) > hbtors and \
                   ANGLE(da1, d1, acp) > 90.0 and \
                   ANGLE(da2, d1, acp) > 90.0:
                    dist = d1.distance(acp)
                    if dist > hbd:
                        hbene = hbene * (hbdmax - dist)/(hbdmax - hbd)
                    ehb = ehb + hbene
                    break
    return -ehb

def print_hbond_score(mol, hblist, file, ABS=abs, ANGLE=Atom3d.angle,
                     TORSION=Atom3d.torsion):
    """List details of the hydrogen bond score to a file

    Arguments

        o *mol* - instance of LinusMol - molecule whose hydrogen bond
        energy is to be computed

        o *hblist* - object - the object returned by the function
        *make_hbond_list* in the *construct* module

        o *file* - string or file object - name of a file or a file
        previously opened for writing.

    Result

        The hydrogen bond energy as a floating point number
    """    
    if type(file) is type(''):
        ff = open(file, 'wb')
    else:
        ff = file

    Write = ff.write
    Write('\n------ Hydrogen Bond Score Listing -----\n')

    elist = []
    ehb = 0.0
    for dlist, alist in hblist:
        d1, da1, da2 = dlist
        Close = d1.close
        for acp, hbd, hbdmax, hbtors, hbene in alist:
            if Close(acp, hbdmax) and \
               ANGLE(da1, d1, acp) > 90.0 and \
               ANGLE(da2, d1, acp) > 90.0 and \
               ABS(TORSION(acp, d1, da1, da2)) > hbtors:
                dist = d1.distance(acp)
                if dist > hbd:
                    hbene = hbene * (hbdmax - dist)/(hbdmax - hbd)
                ehb = ehb + hbene
                elist.append((d1, acp, dist, hbene))
                break

    rnames = mol.residue_names
    atoms = mol.atoms
    fai = mol.residue_first_atom_indices

    for rec in elist:
        d1, acp, dist, hbene = rec
        Write('%4d %3s %4s %4d %3s %4s Distance = %.3f Score = %.3f\n' %
              (d1.resnum+1, rnames[d1.resnum], d1.name, 
               acp.resnum+1, rnames[acp.resnum], acp.name, dist, hbene))
    Write('\nTotal Hydrogen Bond Score = %.3f\n' % -ehb)
    if ff <> file:
        ff.close()
    return -ehb

def solvation_score(atoms, wmax):
    """Compute the solvation score of a collection of atoms

    Arguments

        o *atoms* - list - a list of Atom3d instances

        o *wmax* - integer - maximum sequence separation between
        pairs of residues 

    Result

        The accessible surface area and solvation score as
        floating point numbers
    """    
    raise AttributeError, 'Function Not Implemented'

def print_solvation_score(p, cavpar, wmax, file):
    """List the solvation score of a molecule to file

    Arguments

        o *p* - instance of LinusMol - molecule whose solvation score
        is to be computed

        o *cavpar* - float - the cavitation parameter

        o *wmax* - integer - maximum sequence separation between a pair of
        residues

        o *file* - string or file object - name of a file or a file
        previously opened for writing.

    Result

        The accessible surface area and solvation score as
        floating point numbers
    """
    
    if type(file) is type(''):
        ff = open(file, 'wb')
    else:
        ff = file
    Write = ff.write
    Write('\n------ Solvation Score Listing -----\n')
    atoms = p.satoms
    area, esol = solvation_score(atoms, wmax)
    Write('Total Accessible Area = %g\n' % area)
    cavscr = area * cavpar
    Write('Cavitation Term = %g\n' % cavscr)
    for atom in atoms:
        Write('%s %6d Area=%12.6f Score=%12.6f\n' % (atom.name,
                                                     atom.resnum,
                                                     atom.sasa,
                                                     atom.sasa*atom.sasapar))
    s = cavscr + esol
    Write('Total Solvation Score = %g\n' % s)
    return s

   
try:    
    from linusc import *
except ImportError:
    print 'Using slower Python version of linusScore'
    
