"""Functions for reading and writing PDB files.

Functions in this module

    o *protein_from_pdb* - instance of LinusMol - should never
    have to be called directly by the user

    o *write_pdb* - write instance of LinusMol to a file in
    PDB format
"""

import Atom3d
from linusRes import *
StringType = type('')

def get_residue_atoms(res_name):
    res = RESIDUES[res_name]
    atoms = []
    for arec in res:
        a = Atom3d.Atom3d()
        a.name = arec[0]
        a.radius = arec[4]
        a.bsep0 = arec[-2]
        a.bsep1 = arec[-1]
        atoms.append(a)
    return atoms

def get_atom_with_name(atoms, start, end, name):
    for i in range(start, end):
        if atoms[i].name == name: return atoms[i]
    return None

def protein_from_pdb(p):
    """Fill in the attributes of a LinusMol

    Arguments

        o *p* - instance of LinusMol

    Result

        None
    """    
    from string import strip
    data = open(p.filename, 'rb').readlines()
    na = nr = -1
    a = p.atoms = []
    r = p.residue_names = []
    f = p.residue_first_atom_indices = []

    current_res = None

    for line in data:
        if line[:4] == 'ATOM':
            na = na + 1
            rnam = line[17:20]
            rnum = line[22:27]
            if rnum <> current_res:
                nr = nr + 1
                f.append(na)
                r.append(rnam)
                current_res = rnum
                a.extend(get_residue_atoms(rnam))
                start = na
                end = len(a)
            aname = line[12:16]
            atom = get_atom_with_name(a, start, end, aname)
            if atom is None:
                raise NameError, 'Unknown atom "%s" in resiude %s' % (aname, rnam)
            atom.x = float(line[30:38])
            atom.y = float(line[38:46])
            atom.z = float(line[46:54])
            atom.resnum = nr
        elif line[:4] == 'COMP':
            p.name = strip(line[10:70])
        elif line[:3] in ('TER', 'END'):
            break
    p.num_atoms = na + 1
    nr = p.num_residues = nr + 1
    f.append(p.num_atoms)

    # create and initialize Zmatrix

    for i in range(nr):
        res = RESIDUES[r[i]]
        for atom_record in res:
            atom = get_atom_with_name(a, f[i], f[i+1],
                                      atom_record[0])
            fp, fpname = atom_record[1]
            fp = fp + i
            if fp >= 0 and fp < nr:
                atom.first_parent = get_atom_with_name(a, f[fp],
                                                       f[fp+1], fpname)
                atom.fp_distance = atom.distance(atom.first_parent)
            fp, fpname = atom_record[2]
            fp = fp + i
            if fp >= 0 and fp < nr:
                atom.second_parent = get_atom_with_name(a, f[fp],
                                                       f[fp+1], fpname)
                atom.sp_angle = Atom3d.angle(atom, atom.first_parent,
                                             atom.second_parent)
            fp, fpname = atom_record[3]
            fp = fp + i
            if fp >= 0 and fp < nr:
                atom.third_parent = get_atom_with_name(a, f[fp],
                                                       f[fp+1], fpname)
                atom.tp_torsion = Atom3d.torsion(atom, atom.first_parent,
                                                 atom.second_parent,
                                                 atom.third_parent)

    Atom3d.commit_internal_coords(a)

    # setup pointers to the phi, psi, omega and chi torsions for each
    # residue

    p.phi_atoms = [None]*nr
    p.psi_atoms = [None]*nr
    p.omega_atoms = [None]*nr
    p.chi_atoms = [None]*nr
    
            
    for i in range(p.num_residues):
        name = r[i]
        try:
            phiatom = PhiAtoms[name]
        except KeyError:
            phiatom = PhiAtoms['ANY']

        try:
            psiatom = PsiAtoms[name]
        except KeyError:
            psiatom = PsiAtoms['ANY']

        try:
            omeatom = OmeAtoms[name]
        except KeyError:
            omeatom = OmeAtoms['ANY']

        chiatom = ChiAtoms.get(name, None)

        start = f[i]
        end = f[i+1]

        atom = get_atom_with_name(a, start, end, phiatom)
        if atom: p.phi_atoms[i] = atom

        atom = get_atom_with_name(a, start, end, psiatom)
        if atom: p.psi_atoms[i] = atom

        atom = get_atom_with_name(a, start, end, omeatom)
        if atom: p.omega_atoms[i] = atom

        if chiatom:
            chis = ()
            for atomname in chiatom:
                chis = chis + (get_atom_with_name(a, start, end, atomname),)
            p.chi_atoms[i] = chis
            

def write_pdb(p, filename, title=''):
    """Write out an instance of LinusMol to file in PDB format

    Arguments

         o *p* - instance of LinusMol

         o *filename* - string or file opened for writing - file to
         which molecule will be written

         o *title* - string - title for the molecule
    """
    
    if type(filename) is StringType:
        ff = open(filename, 'wb')
    else:
        ff = filename

    write = ff.write

    write('COMPND     %s\n' % title)

    pdbfmt = 'ATOM  %5d %-4s %-3s  %4d    %8.3f%8.3f%8.3f  0.00  0.00\n'

    fai = p.residue_first_atom_indices
    atoms = p.atoms
    residue_names = p.residue_names

    for i in xrange(p.num_residues):
        rname = residue_names[i]
        start = fai[i]
        end = fai[i+1]
        for j in range(start, end):
            atom = atoms[j]
            write(pdbfmt % (j+1, atom.name, rname, i+1, atom.x, atom.y,
                            atom.z))

    write('TER\n')

    if ff <> filename:
        write('END\n')
        ff.close()
