"""Script to process the structures written out during a LINUS simulation
and list for each pair of residues the fraction of the structures in
which they were in contact.  Residues which were never in contact 
are not listed.  The output is written to stdout and consists of
3 columns:

   Column 1 - Residue Number
   Column 2 - Residue Number
   Column 3 - Fraction of structures in which the pair of residues are
              in contact
"""

USAGE = """
python concnt.py pdbfile outfile winmin wimax probe

  pdbfile = name of pdb file that was input to LINUS simulation
  outfile = name of file to which sampled structures were written
            during the course of the simulation
  winmin = minimum sequence separation between pairs of residues
           that are allowed to be in contact
  winmax = maximum sequence separation between pairs of residues
           that are allowed to be in contact 
  probe = distance overwhich the contact energy for a pair of atoms
          is linearly scaled to zero
"""
            
from pylinus import linusSim, construct
import sys

def contact_matrix(clist, cprobe):
    cdict = {}
    eborpc = 1.0/cprobe
    econ = 0.0
    for c1, c2, dmin, ene in clist:
        dmax = dmin + cprobe
        if c1.close(c2, dmax):
            d12 = c1.distance(c2)
            if not c1.first_parent.close(c2.first_parent, d12):
                cdict[(c1.resnum, c2.resnum)] = 1
    return cdict.keys()

def get_conformation(out_file, protein):
    nextline = out_file.readline
    atoms = protein.atoms
    Float = float
    while 1:
        line = nextline()
        if not line:
            break
        if line[:6] == 'COMPND':
            break
    if not line:
        return 0
    for atom in atoms:
        line = nextline()
        atom.x, atom.y, atom.z = map(Float, (line[30:38],
                                             line[38:46],
                                             line[46:54]))
    return 1
    
def concnt(input_file, out_file, winmin, winmax, cprobe):
    prot = linusSim.Linus(input_file, None).protein
    clist = construct.make_contact_list(prot, winmin, winmax)
    out = open(out_file, 'rb')
    c_dict = {}
    n = 0
    while 1:
        ok = get_conformation(out, prot)
        if not ok: break
        d = contact_matrix(clist, cprobe)
        for key in d:
            c_dict[key] = c_dict.get(key, 0) + 1
        n = n + 1

    n = float(n)
    keys = c_dict.keys()
    keys.sort()
    for i in xrange(len(keys)):
        key = keys[i]
        keys[i] = (key, c_dict[key]/n)
    return keys

if __name__ == "__main__":
    import sys

    try:
        input_file, output_file, winmin, winmax, cprobe = sys.argv[1:]
    except:
        print USAGE
        sys.exit()
            
    result = concnt(input_file, output_file, int(winmin), int(winmax),
                    float(cprobe))

    for item, value in result:
        print '%4d %4d %5.3f' % (item[0], item[1], value)

 
