#! /usr/bin/env python
import string
import math

#
# try and import Numeric and associated functions. If Numeric is not
# available we define the necessary functions in python - which
# will be a bit slower
#

try:
    import Numeric
except ImportError:
    def CUMSUM(a):
        l = [a[0]]
        for i in xrange(1, len(a)):
            l.append(a[i] + l[-1])
        return l
    def ZEROS(n, dummy):
        return [0.0]*n
    def ARRAY(v, dummy):
        return list(v)
else:
    CUMSUM = Numeric.add.accumulate
    ZEROS = Numeric.zeros
    ARRAY = Numeric.array

#
# the following procedure is used to construct coordinates for an
# atom given a Z-Matrix description.  This function is used only
# if the faster vecmod module is not present
#
def int2vec(v1, dis, v2, ang, v3, tors, sin=math.sin,\
            cos=math.cos, sqrt=math.sqrt):
    """int2vec(v1, dis, v2, ang, v3, tors)
        Compute coordinates for vector v, given its 'dis'tance
        from vector v1, 'ang'le v-v1-v2 and 'tors'ion v-v1-v2-v3.'
        Both ang and tors should be in degrees.
    """

    DEG2RAD = math.pi/180
    ang = ang*DEG2RAD
    sina = sin(ang)
    cosa = cos(ang)

    tors = (tors*DEG2RAD)    
    sint = sina * sin(tors)
    cost = sina * cos(tors)

    x1, y1, z1 = v1
    x2, y2, z2 = v2
    x3, y3, z3 = v3
    
    u1x = x2 - x3
    u1y = y2 - y3
    u1z = z2 - z3
    d = 1.0 / sqrt(u1x*u1x + u1y*u1y + u1z*u1z)
    u1x = u1x*d
    u1y = u1y*d
    u1z = u1z*d

    u2x = x1 - x2
    u2y = y1 - y2
    u2z = z1 - z2;
    d = 1.0 / sqrt(u2x*u2x + u2y*u2y + u2z*u2z)
    u2x = u2x*d
    u2y = u2y*d
    u2z = u2z*d

    cosine = u1x*u2x + u1y*u2y + u1z*u2z

    if (abs(cosine) < 1.0):
        sine = 1.0/sqrt(1.0 - cosine*cosine)
    else:
        sine = 1.0/sqrt(cosine*cosine - 1.0)
            
    u3x = sine * (u1y*u2z - u1z*u2y)
    u3y = sine * (u1z*u2x - u1x*u2z)
    u3z = sine * (u1x*u2y - u1y*u2x)

    u4x = cost * (u3y*u2z - u3z*u2y)
    u4y = cost * (u3z*u2x - u3x*u2z)
    u4z = cost * (u3x*u2y - u3y*u2x)

    return [x1 + dis*(-u2x*cosa + u4x + u3x*sint),
            y1 + dis*(-u2y*cosa + u4y + u3y*sint),
            z1 + dis*(-u2z*cosa + u4z + u3z*sint)]

try:
    from vecmod import int2vec
except ImportError:
    pass

#
# some local classes for passing around data - the class Atom stores
# an atom description as read from the parameter file and Residue
# is an ordered collection of Atom instances with additional information
# read from the parameter file
#
class Atom:
    def __init__(self, name, p1, p2, p3, d1, d2, d3, iscore, chitype):
        self.name = name
        self.p1 = p1        # first parent
        self.p2 = p2        # second parent
        self.p3 = p3        # third parent
        self.d1 = d1        # distance from first parent
        self.d2 = d2        # angle made with first and second parents
        self.d3 = d3        # torsion made withe first, second and third parents
        self.iscore = (iscore == '+' or iscore == 1)
        self.chitype = chitype

    def copy(self):
        return Atom(self.name, self.p1, self.p2, self.p3, self.d1,
                    self.d2, self.d3, self.iscore, self.chitype)


class Residue:
    def __init__(self, name, atoms=None):
        self.name = name
        if not atoms is None:
            self.atoms = atoms
        else:
            self.atoms = []
    def add_atom(self, atom):
        self.atoms.append(atom)

    def __getitem__(self, i):
        return self.atoms[i]

    def __len__(self):
        return len(self.atoms)

    def chiatom(self, chiname):
        for atom in self.atoms:
            if atom.chitype == chiname:
                return atom

    def copy(self):
        newRes = Residue(self.name)
        for atom in self:
            newRes.add_atom(atom.copy())
        return newRes


#
# helper functions
#
def _build_atom(line, split=string.split, atof=string.atof, atoi=string.atoi,
                       upper=string.upper, Atom=Atom):
    args = split(upper(line))
    return Atom(args[0], atoi(args[4]), atoi(args[5]), atoi(args[6]),
                atof(args[1]), atof(args[2]), atof(args[3]), args[7],
                args[8])

def _read_parm_file(file):
    if type(file) == type(''):
        ff = open(file, 'r')
    else:
        ff = file
    nextline = ff.readline
    Strip = string.strip
    Split = string.split
    Upper = string.upper
    Int = string.atoi

    Residues = {}
    while 1:
        line = nextline()
        if not line:
            break
        line = Strip(line)
        line = Upper(line)
        if line[:4] <> 'NAME':
            continue
        line = Split(line)
        name = line[1]
        numatm = Int(line[-1])
        newRes = Residue(name)
        for i in range(numatm):
            line = nextline()
            newRes.add_atom(_build_atom(line))
        Residues[name] = newRes
    return Residues


def _normalize_indices(rl):
    
    rp = map(len, rl)
    rp.insert(0, 0)
    rp = CUMSUM(rp)
    for i in range(len(rl)):
        res = rl[i]
        for atom in res:
            p = atom.p1
            if p < 0:
                atom.p1 = rp[i-1] + abs(p) - 1
            else:
                atom.p1 = rp[i] + p - 1
            p = atom.p2
            if p < 0:
                atom.p2 = rp[i-1] + abs(p) - 1
            else:
                atom.p2 = rp[i] + p - 1
            p = atom.p3
            if p < 0:
                atom.p3 = rp[i-1] + abs(p) - 1
            else:
                atom.p3 = rp[i] + p - 1

        
def build_peptide(rl):
    _normalize_indices(rl)
    dtor = math.pi/180.0
    res0 = rl[0]
    res0[0].xyz = ZEROS(3, 'd')
    a2 = res0[1]
    res0[1].xyz = ARRAY((a2.d1, 0, 0), 'd')
    a3 = res0[2]
    if a3.p1 == 1:
        x = a2.d1 - a3.d1*math.cos(dtor*a3.d2)
    else:
        x = a3.d1*math.cos(dtor*a3.d2)
    y = a3.d1 * math.sin(a3.d2*dtor)
    res0[2].xyz = ARRAY((x, y, 0.0), 'd')
    atomlist = []
    for res in rl:
        for atom in res.atoms:
            atomlist.append(atom)

    for i in xrange(3, len(atomlist)):
        a = atomlist[i]
        p1 = atomlist[a.p1].xyz
        p2 = atomlist[a.p2].xyz
        p3 = atomlist[a.p3].xyz
        a.xyz = int2vec(p1, a.d1, p2, a.d2, p3, a.d3)



def _read_input(file):
    residues = []
    if type(file) == type(''):
        ff = open(file, 'r')
    else:
        ff = file
    nextline = ff.readline
    allh = 0
    title = ''
    defphi = None
    defpsi = None
    Strip = string.strip
    Split = string.split
    Upper = string.upper
    Float = string.atof
    while 1:
        line = nextline()
        if not line:
            break
        line = Strip(line)
        if not line or line[:1] == '#': continue
        line = Upper(line)
        if line[:3] == 'RES':
            sline = Split(line)
            name = sline[1]
            args = sline[2:]
            res = {'name': name}
            if args:
                for i in range(0, len(args), 2):
                    res[args[i]] = Float(args[i+1])
            residues.append(res)
        elif line[:5] == 'TITLE':
            title = Split(line)[1]
        elif line[:3] == 'DEF':
            sline = Split(line)
            ang = sline[1]
            val = Float(sline[2])
            if ang == 'PHI':
                defphi = val
            elif ang == 'PSI':
                defpsi = val
        elif line[:4] == 'ALLH':
            allh = 1
        else:
            raise ValueError,'Unknown keyword in input file: %s' % Split(line)[0]

    for res in residues:
        if not res.has_key('PHI'):
            if defphi:
                res['PHI'] = defphi
    if defpsi:
        defpsi = defpsi + 180.0
        if defpsi > 180:
            defpsi = defpsi - 360.0
    for res in residues:
        if not res.has_key('PSI'):
            if defpsi: res['PSI'] = defpsi
        else:
            val = res['PSI'] + 180.0
            if val > 180:
                res['PSI'] = val - 360.
            else:
                res['PSI'] = val
    return title, allh, residues

def make_chain(rl, db):
    chain = []
    for res in rl:
        name = res['name']
        if not db.has_key(name):
            raise TypeError, 'Unknown Residue: %s' % name
        Res = db[name].copy()
        keys = res.keys()
        keys.remove('name')
        for key in keys:
            val = res[key]
            for atom in Res.atoms:
                if atom.chitype == key:
                    atom.d3 = val
        chain.append(Res)
    return chain

def write_pdb(title, chain, file, allh):
    if type(file) == type(''):
        f = open(file, 'w')
    else:
        f = file

    fmt1 = 'ATOM  %5d  %-4s%-4s %4d    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
    fmt2 = 'ATOM  %5d %-4s %-4s %4d    %8.3f%8.3f%8.3f%6.2f%6.2f\n'
    f.write('COMPND    %s\n' % title)
    seq = []
    nres = len(chain)
    line_num = 1
    for i in xrange(0, nres, 13):
        prefix = 'SEQRES%4d  %4d  ' % (line_num, nres)
        rec = [prefix]
        for j in range(i, min(i+13, nres)):
            rec.append(' '+chain[j].name)
        rec.append('\n')
        f.write(string.join(rec, ''))
        line_num = line_num + 1
    i = 1
    j = 1
    for res in chain:
        res_name = res.name
        for atom in res:
            name = atom.name
            if name == 'DU': continue
            if atom.iscore or allh:
                if name[:1] in '123':
                    fmt = fmt2
                else:
                    fmt = fmt1
                f.write(fmt % ( i, name, res_name, j, atom.xyz[0], atom.xyz[1],
                                atom.xyz[2], 1.0, 0.0))
                i = i + 1
        j = j + 1
    f.write('END\n')
    f.close()

        
def run(in_file, parm_file, out_file):
    title, allh, residues = _read_input(in_file)
    db = _read_parm_file(parm_file)
    chain = make_chain(residues, db)
    build_peptide(chain)
    write_pdb(title, chain, out_file, allh)

if __name__ == "__main__":
    import sys, getopt, os
    import pylinus_1_0.utils
    
    parm_file = os.path.dirname(pylinus_1_0.utils.__file__) + os.sep + 'ribosome.dat'
    in_file = out_file = None

    args = getopt.getopt(sys.argv[1:], 'i:p:o:')[0]
    for opt, val in args:
        if opt == '-i':
            in_file = val
        elif opt == '-p':
            parm_file = val
        elif opt == '-o':
            out_file = val

    if in_file is None:
        in_file = sys.stdin
        ##in_file = raw_input('Enter Name of input file: ')
        ##if not in_file:
        ##    raise ValueError, 'Input file not specified'

    if out_file is None:
        out_file = sys.stdout
        ##out_file = raw_input('Enter Name of output file: ')
        ##if not out_file:
        ##    raise ValueError, 'Output file not specified'

    run(in_file, parm_file, out_file)

        
            
    
        
