import copy

HAVE_POP = hasattr([], 'pop')
HAVE_EXTEND = hasattr([], 'extend')

class TypedList:
    """A Python list restricted to having objects of the same type.
    An instance of a TypedList is created as follows:
    
    mylist = TypedList(function, [elements])
    
    where function is a python function which takes an argument
    and returns 1 or 0 indicating whether the object represented
    by the argument is of the correct type, and elements is an optional
    list of elements to be added into the instance.  Here is a
    full blown example.

    def is_int(o):
        return type(o) == type(0)

    mylist = TypedList(is_int, [0, 1, 2, 3])

    New elements are added to the list as follows:
    mylist.append(25)

    Instances of TypedList support all operations available for
    Python Lists (as of Python version 1.5.2a2)
    """

    _is_a_typed_list = 1
    
    def __init__(self, function, elements=None):
        self._func = function
        self.elements = []
        if not elements is None:
            if self._func(elements):
                self.elements.append(elements)
            else:
                for el in elements:
                    self.append(el)

    def append(self, el):
        if self._func(el):
            self.elements.append(el)
        else:
            raise TypeError, 'Element to be added to list has incorrect type.'

    def __len__(self): return len(self.elements)

    def __repr__(self):
        return "%s(%s, %s)" % (self.__class__, self._func.__name__,
                              `self.elements`)

    def __str__(self):
        return `self.elements`
    
    def __getitem__(self, i): return self.elements[i]
    
    def __setitem__(self, i, v):
        if self._func(v):
            self.elements[i] = v
        else:
            raise TypeError, 'Item not of correct type in __setitem__'

    def __delitem__(self, i): del self.elements[i]

    def __getslice__(self, i, j):
        new = self.__class__(self._func)
        new.elements = self.elements[i:j]
        return new

    def __setslice__(self, i, j, v):
        if self._alltrue(v):
            self.elements[i:j] = v

    def __delslice__(self, i, j):
        del self.elements[i:j]

    def __add__(self, other):
        if not hasattr(other, '_is_a_typed_list'):
            raise TypeError,'List to be concatenated not instance of %s' %\
                  self.__class__
        if self._func <> other._func:
            raise TypeError, 'Lists to be added not of same type'
        new = self.__class__(self._func)
        new.elements = self.elements + other.elements
        return new

    def __mul__(self, other):
        if type(other) == type(0):
            new = self.__class__(self._func)
            new.elements = self.elements * other
            return new
        else:
            raise TypeError, "can't multiply list with non-int"
        
    __rmul__ = __mul__

    def __copy__(self):
        new = self.__class__(self._func)
        for el in self.elements:
            new.elements.append(el.__copy__())
        have = new.__dict__.has_key
        for key in self.__dict__.keys():
            if not have(key):
                new.__dict__[key] = copy.deepcopy(self.__dict__[key])
        return new

    __deepcopy__ = clone = __copy__

    def _alltrue(self, els):
        return len(els) == len(filter(None, map(self._func, els)))

    def sort(self): self.elements.sort()

    def reverse(self): self.elements.reverse()
    
    def count(self, el): return self.elements.count(el)
    
    def extend(self, els):
        if self._alltrue(els):
            if HAVE_EXTEND:
                self.elements.extend(els)
            else:
                for el in els:
                    self.elements.append(el)
        else:
            raise TypeError, 'One or more elements of list not of correct type'
                    
    def pop(self):
        if HAVE_POP:
            return self.elements.pop()
        else:
            el = self.elements[-1]
            del self.elements[-1]
            return el
    
    def index(self, el): return self.elements.index(el)
    
    def remove(self, el): self.elements.remove(el)
    
    def insert(self, pos, el):
        if self._func(el):
            self.elements.insert(pos, el)
        else:
            raise TypeError, 'Item not of correct type in insert'

    def indices(self, len=len):
        return xrange(len(self.elements))

    def reverse_indices(self, len=len):
        return xrange(len(self.elements)-1, -1, -1)

    
    
        
        
